@@ -1,7 +1,7 @@
 /*
  * monorail.c
  *
- * Copyright (C) 2009 Collabora Ltd. <http://www.collabora.co.uk/>
+ * Copyright (C) 2009-2010 Collabora Ltd. <http://www.collabora.co.uk/>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -18,6 +18,8 @@
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
+#include <glib.h>
+
 #include <telepathy-glib/debug.h>
 
 #include <unique/unique.h>
@@ -31,6 +33,28 @@
 #include "monorail.h"
 #include "transfer.h"
 
+/* Monorail can send files either when asked from its main window or when it
+ * receives a list of files to send from its D-Bus Send interface. Any of the
+ * those two methods will use a MonorailSendContext to store the context.
+ */
+typedef struct
+{
+  enum {
+    CONTEXT_MAIN,
+    CONTEXT_DBUS
+  } context_type;
+  GtkWidget *dialog;
+  GtkWidget *contact_view;
+  GtkWidget *live_search;
+  GtkWidget *file_window;
+  GFile *selected_file;
+  /* list of GFile */
+  GSList *selected_files;
+  OssoABookContact *selected_master_contact;
+  OssoABookContact *selected_contact;
+  MonorailData *monorail_data;
+} MonorailSendContext;
+
 static gboolean
 no_transfers_timeout (gpointer data)
 {
@@ -130,6 +154,17 @@ main_window_delete_event (GtkWidget *widget,
   return TRUE;
 }
 
+static gboolean
+dialog_delete_event (GtkWidget *widget,
+    GdkEvent *event,
+    MonorailSendContext *data)
+{
+  monorail_transfers_dec (data->monorail_data, FALSE);
+
+  return FALSE;
+}
+
+
 static void
 transfers_button_clicked (GtkButton *button,
     MonorailData *data)
@@ -158,12 +193,13 @@ create_menu (MonorailData *data)
 }
 
 static void
-send_file (MonorailData *data)
+send_file (MonorailSendContext *data)
 {
   TpDBusDaemon *dbus;
   TpAccount *account;
   EmpathyFTFactory *factory;
   McAccount *mcaccount;
+  GSList *files;
 
   if (data->selected_contact == NULL
       || !osso_abook_contact_is_roster_contact (data->selected_contact))
@@ -178,14 +214,30 @@ send_file (MonorailData *data)
   account = tp_account_new (dbus, tp_proxy_get_object_path (mcaccount), NULL);
   g_object_unref (dbus);
 
-  factory = monorail_handler_get_factory (data->handler);
+  factory = monorail_handler_get_factory (data->monorail_data->handler);
 
   DEBUG ("Starting new outgoing transfer to %s",
       osso_abook_contact_get_bound_name (data->selected_contact));
 
-  empathy_ft_factory_new_transfer_outgoing (factory, account,
-      osso_abook_contact_get_bound_name (data->selected_contact),
-      data->selected_file, monorail_handler_get_name (data->handler));
+  if (data->context_type == CONTEXT_MAIN)
+    {
+      empathy_ft_factory_new_transfer_outgoing (factory, account,
+          osso_abook_contact_get_bound_name (data->selected_contact),
+          data->selected_file,
+          monorail_handler_get_name (data->monorail_data->handler));
+    }
+  else
+    {
+      files = data->selected_files;
+      while (files != NULL)
+        {
+          empathy_ft_factory_new_transfer_outgoing (factory, account,
+              osso_abook_contact_get_bound_name (data->selected_contact),
+              files->data,
+              monorail_handler_get_name (data->monorail_data->handler));
+          files = files->next;
+        }
+    }
 
   g_object_unref (account);
 }
@@ -193,29 +245,121 @@ send_file (MonorailData *data)
 static void
 confirm_dialog_response_cb (GtkDialog *dialog,
     gint response_id,
-    MonorailData *data)
+    MonorailSendContext *data)
 {
   if (response_id != GTK_RESPONSE_YES)
     goto out;
 
   send_file (data);
 
-  gtk_widget_destroy (data->file_window);
-  data->file_window = NULL;
+  if (data->context_type == CONTEXT_MAIN)
+    {
+      gtk_widget_destroy (data->file_window);
+      data->file_window = NULL;
+    }
+  else
+    {
+      monorail_transfers_dec (data->monorail_data, FALSE);
+      gtk_widget_destroy (data->dialog);
+      data->dialog = NULL;
+
+      if (data->selected_master_contact != NULL)
+        {
+          g_object_unref (data->selected_master_contact);
+          data->selected_master_contact = NULL;
+        }
+
+      if (data->selected_contact != NULL)
+        {
+          g_object_unref (data->selected_contact);
+          data->selected_contact = NULL;
+        }
+
+      if (data->selected_files != NULL)
+        {
+          g_slist_foreach (data->selected_files, (GFunc) g_object_unref, NULL);
+          g_slist_free (data->selected_files);
+          data->selected_files = NULL;
+        }
+
+      data->monorail_data = NULL;
+      g_slice_free (MonorailSendContext, data);
+    }
 
 out:
   gtk_widget_destroy (GTK_WIDGET (dialog));
 }
 
 static void
-file_activated_cb (HildonFileSelection *self,
-    MonorailData *data)
+ask_to_send_file (MonorailSendContext *data)
 {
   GtkWidget *dialog, *label;
-  GSList *selection;
-  gchar *basename;
   gchar *text;
 
+  dialog = gtk_dialog_new ();
+  gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
+  gtk_window_set_destroy_with_parent (GTK_WINDOW (dialog), TRUE);
+
+  if (data->context_type == CONTEXT_MAIN)
+    {
+      gtk_window_set_transient_for (GTK_WINDOW (dialog),
+          GTK_WINDOW (data->file_window));
+    }
+
+  gtk_widget_show (gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_YES,
+          GTK_RESPONSE_YES));
+  gtk_widget_show (gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_NO,
+          GTK_RESPONSE_NO));
+
+  if (data->context_type == CONTEXT_MAIN)
+    {
+      gchar *basename;
+      basename = g_file_get_basename (data->selected_file);
+      text = g_strdup_printf ("Send %s\nto %s?", basename,
+          osso_abook_contact_get_display_name (data->selected_master_contact));
+      g_free (basename);
+      gtk_window_set_title (GTK_WINDOW (dialog), "Send file?");
+    } 
+  else
+   {
+      guint count = g_slist_length (data->selected_files);
+      if (count == 1)
+        {
+          gchar *basename;
+          basename = g_file_get_basename (data->selected_files->data);
+          text = g_strdup_printf ("Send %s\nto %s?", basename,
+              osso_abook_contact_get_display_name (data->selected_master_contact));
+          g_free (basename);
+          gtk_window_set_title (GTK_WINDOW (dialog), "Send file?");
+        }
+      else
+        {
+          text = g_strdup_printf ("Send %u files to %s?",
+              g_slist_length (data->selected_files),
+              osso_abook_contact_get_display_name (data->selected_master_contact));
+          gtk_window_set_title (GTK_WINDOW (dialog), "Send files?");
+        }
+   }
+
+  label = gtk_label_new (text);
+  gtk_widget_show (label);
+  gtk_box_pack_start (
+      GTK_BOX (gtk_dialog_get_content_area (GTK_DIALOG (dialog))), label, TRUE, TRUE, 0);
+
+  g_free (text);
+
+  g_signal_connect (dialog, "response",
+      G_CALLBACK (confirm_dialog_response_cb), data);
+
+  gtk_widget_show (dialog);
+}
+
+static void
+file_activated_cb (HildonFileSelection *self,
+    MonorailSendContext *data)
+{
+  GSList *selection;
+
   selection = hildon_file_selection_get_selected_uris (self);
 
   if (data->selected_file != NULL)
@@ -230,39 +374,11 @@ file_activated_cb (HildonFileSelection *self,
   g_slist_foreach (selection, (GFunc) g_free, NULL);
   g_slist_free (selection);
 
-  basename = g_file_get_basename (data->selected_file);
-
-  dialog = gtk_dialog_new ();
-  gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
-  gtk_window_set_destroy_with_parent (GTK_WINDOW (dialog), TRUE);
-  gtk_window_set_title (GTK_WINDOW (dialog), "Send file?");
-  gtk_window_set_transient_for (GTK_WINDOW (dialog),
-      GTK_WINDOW (data->file_window));
-
-  gtk_widget_show (gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_YES,
-          GTK_RESPONSE_YES));
-  gtk_widget_show (gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_NO,
-          GTK_RESPONSE_NO));
-
-  text = g_strdup_printf ("Send %s\nto %s?", basename,
-      osso_abook_contact_get_display_name (data->selected_master_contact));
-
-  label = gtk_label_new (text);
-  gtk_widget_show (label);
-  gtk_box_pack_start (
-      GTK_BOX (gtk_dialog_get_content_area (GTK_DIALOG (dialog))), label, TRUE, TRUE, 0);
-
-  g_free (basename);
-  g_free (text);
-
-  g_signal_connect (dialog, "response",
-      G_CALLBACK (confirm_dialog_response_cb), data);
-
-  gtk_widget_show (dialog);
+  ask_to_send_file(data);
 }
 
 static void
-show_file_manager (MonorailData *data)
+show_file_manager (MonorailSendContext *data)
 {
   GtkWidget *widget;
   HildonFileSystemModel *model;
@@ -270,7 +386,7 @@ show_file_manager (MonorailData *data)
   /* Create new window */
   data->file_window = hildon_stackable_window_new ();
   hildon_window_set_app_menu (HILDON_WINDOW (data->file_window),
-      create_menu (data));
+      create_menu (data->monorail_data));
   gtk_window_set_title (GTK_WINDOW (data->file_window), "Select a file");
 
   model = g_object_new (HILDON_TYPE_FILE_SYSTEM_MODEL,
@@ -316,7 +432,7 @@ show_file_manager (MonorailData *data)
 static gboolean
 hide_live_search (gpointer user_data)
 {
-  MonorailData *data = user_data;
+  MonorailSendContext *data = user_data;
 
   gtk_widget_hide (data->live_search);
 
@@ -325,7 +441,7 @@ hide_live_search (gpointer user_data)
 
 typedef struct
 {
-  MonorailData *monorail_data;
+  MonorailSendContext *data;
   GtkWidget *dialog;
   GHashTable *button_to_contact;
   OssoABookContact *master_contact;
@@ -346,7 +462,7 @@ static void
 roster_contact_button_clicked_cb (GtkButton *button,
     RosterContactChooserData *chooser_data)
 {
-  MonorailData *data = chooser_data->monorail_data;
+  MonorailSendContext *data = chooser_data->data;
   OssoABookContact *contact;
 
   contact = g_hash_table_lookup (chooser_data->button_to_contact, button);
@@ -370,12 +486,16 @@ roster_contact_button_clicked_cb (GtkButton *button,
   g_slice_free (RosterContactChooserData, chooser_data);
 
   g_idle_add (hide_live_search, data);
-  show_file_manager (data);
+
+  if (data->context_type == CONTEXT_DBUS)
+    ask_to_send_file (data);
+  else
+    show_file_manager (data);
 }
 
 static void
 show_roster_contact_chooser (OssoABookContact *contact,
-    MonorailData *data)
+    MonorailSendContext *data)
 {
   GtkWidget *dialog;
   GList *roster_contacts, *l;
@@ -385,7 +505,7 @@ show_roster_contact_chooser (OssoABookContact *contact,
   gtk_window_set_title (GTK_WINDOW (dialog), "Choose an account");
 
   chooser_data = g_slice_new0 (RosterContactChooserData);
-  chooser_data->monorail_data = data;
+  chooser_data->data = data;
   chooser_data->dialog = dialog;
   chooser_data->master_contact = g_object_ref (contact);
 
@@ -439,7 +559,7 @@ show_roster_contact_chooser (OssoABookContact *contact,
 static void
 contact_activated_cb (OssoABookContactView *view,
     GtkTreePath *path,
-    MonorailData *data)
+    MonorailSendContext *data)
 {
   GList *selection;
   OssoABookContact *contact, *roster_contact = NULL;
@@ -475,7 +595,11 @@ contact_activated_cb (OssoABookContactView *view,
       data->selected_contact = g_object_ref (roster_contact);
       data->selected_master_contact = g_object_ref (contact);
       g_idle_add (hide_live_search, data);
-      show_file_manager (data);
+
+      if (data->context_type == CONTEXT_DBUS)
+        ask_to_send_file (data);
+      else
+        show_file_manager (data);
     }
   else
     {
@@ -494,18 +618,81 @@ notify_show_transfers (MonorailNotify *notify,
 
 static void
 contact_view_done_cb (MonorailContactChooser *chooser,
-    MonorailData *data)
+    GtkWindow *window)
 {
   hildon_gtk_window_set_progress_indicator (
-      GTK_WINDOW (data->contact_window), 0);
+      GTK_WINDOW (window), 0);
 }
 
 static void
 contact_view_working_cb (MonorailContactChooser *chooser,
-    MonorailData *data)
+    GtkWindow *window)
 {
   hildon_gtk_window_set_progress_indicator (
-      GTK_WINDOW (data->contact_window), 1);
+      GTK_WINDOW (window), 1);
+}
+
+/* Send a list of files as requested from the D-Bus Send interface */
+static void
+send_files_cb (MonorailSend *monorail_send,
+    char **filenames,
+    MonorailData *data)
+{
+  GtkWidget *vbox;
+  GtkTreeView *tree_view;
+
+  MonorailSendContext *dialog_context;
+
+  /* Disable the timeout to void quitting Monorail while the dialog is waiting
+   * for the user */
+  monorail_transfers_inc (data, FALSE);
+
+  dialog_context = g_slice_new0 (MonorailSendContext);
+  dialog_context->context_type = CONTEXT_DBUS;
+  dialog_context->monorail_data = data;
+  dialog_context->selected_contact = NULL;
+
+  dialog_context->selected_files = NULL;
+  while (*filenames != NULL)
+    {
+      dialog_context->selected_files =
+        g_slist_prepend (dialog_context->selected_files,
+            g_file_new_for_uri (*filenames));
+      filenames++;
+    }
+
+  dialog_context->dialog = gtk_dialog_new ();
+  hildon_gtk_window_set_progress_indicator (
+      GTK_WINDOW (dialog_context->dialog), 1);
+
+  vbox = GTK_DIALOG (dialog_context->dialog)->vbox;
+  gtk_window_set_title (GTK_WINDOW (dialog_context->dialog), "Select a contact");
+  g_signal_connect (dialog_context->dialog, "delete-event",
+      G_CALLBACK (dialog_delete_event), dialog_context);
+
+  dialog_context->contact_view = monorail_contact_chooser_new ();
+  gtk_box_pack_start (GTK_BOX (vbox), dialog_context->contact_view, TRUE, TRUE, 0);
+  gtk_widget_show (dialog_context->contact_view);
+
+  g_signal_connect (dialog_context->contact_view, "done",
+      G_CALLBACK (contact_view_done_cb), dialog_context->dialog);
+  g_signal_connect (dialog_context->contact_view, "working",
+      G_CALLBACK (contact_view_working_cb), dialog_context->dialog);
+  g_signal_connect (dialog_context->contact_view, "contact-activated",
+      G_CALLBACK (contact_activated_cb), dialog_context);
+
+  tree_view = osso_abook_tree_view_get_tree_view (
+      OSSO_ABOOK_TREE_VIEW (dialog_context->contact_view));
+
+  dialog_context->live_search = osso_abook_live_search_new_with_filter (
+      osso_abook_tree_view_get_filter_model (
+          OSSO_ABOOK_TREE_VIEW (dialog_context->contact_view)));
+  hildon_live_search_widget_hook (HILDON_LIVE_SEARCH (dialog_context->live_search),
+      dialog_context->dialog, GTK_WIDGET (tree_view));
+  gtk_box_pack_start (GTK_BOX (vbox), dialog_context->live_search, FALSE, FALSE, 0);
+
+  gtk_widget_show_all (dialog_context->dialog);
+  gtk_widget_hide (dialog_context->live_search);
 }
 
 int
@@ -513,6 +700,7 @@ main (int argc,
     char **argv)
 {
   MonorailData *data;
+  MonorailSendContext *main_context;
   GtkWidget *vbox;
   gboolean no_ui = FALSE;
   UniqueApp *unique_app;
@@ -546,10 +734,17 @@ main (int argc,
   data->active_transfers = 0;
   data->quit_timeout_id = 0;
   data->no_ui = no_ui;
-  data->selected_file = NULL;
-  data->selected_contact = NULL;
+
+  main_context = g_slice_new0 (MonorailSendContext);
+  main_context->context_type = CONTEXT_MAIN;
+  main_context->monorail_data = data;
+  main_context->selected_file = NULL;
+  main_context->selected_files = NULL;
+  main_context->selected_contact = NULL;
 
   data->contact_window = hildon_stackable_window_new ();
+  hildon_gtk_window_set_progress_indicator (
+      GTK_WINDOW (data->contact_window), 1);
   hildon_window_set_app_menu (HILDON_WINDOW (data->contact_window),
       create_menu (data));
   gtk_window_set_title (GTK_WINDOW (data->contact_window), "Select a contact");
@@ -574,30 +769,37 @@ main (int argc,
   g_signal_connect (data->notify, "show-transfers",
       G_CALLBACK (notify_show_transfers), data);
 
+  data->send = monorail_send_new ();
+
+  g_signal_connect (data->send, "send-files",
+      G_CALLBACK (send_files_cb), data);
+
   vbox = gtk_vbox_new (FALSE, 0);
   gtk_container_set_border_width (GTK_CONTAINER (vbox), 0);
 
-  data->contact_view = monorail_contact_chooser_new ();
-  gtk_box_pack_start (GTK_BOX (vbox), data->contact_view, TRUE, TRUE, 0);
-  gtk_widget_show (data->contact_view);
+  main_context->contact_view = monorail_contact_chooser_new ();
+  gtk_box_pack_start (GTK_BOX (vbox), main_context->contact_view, TRUE, TRUE, 0);
+  gtk_widget_show (main_context->contact_view);
 
-  g_signal_connect (data->contact_view, "done",
-      G_CALLBACK (contact_view_done_cb), data);
-  g_signal_connect (data->contact_view, "working",
-      G_CALLBACK (contact_view_working_cb), data);
-  g_signal_connect (data->contact_view, "contact-activated",
-      G_CALLBACK (contact_activated_cb), data);
+  g_signal_connect (main_context->contact_view, "done",
+      G_CALLBACK (contact_view_done_cb), data->contact_window);
+  g_signal_connect (main_context->contact_view, "working",
+      G_CALLBACK (contact_view_working_cb), data->contact_window);
+  g_signal_connect (main_context->contact_view, "contact-activated",
+      G_CALLBACK (contact_activated_cb), main_context);
 
   tree_view = osso_abook_tree_view_get_tree_view (
-      OSSO_ABOOK_TREE_VIEW (data->contact_view));
+      OSSO_ABOOK_TREE_VIEW (main_context->contact_view));
 
-  data->live_search = osso_abook_live_search_new_with_filter (
+  main_context->live_search = osso_abook_live_search_new_with_filter (
       osso_abook_tree_view_get_filter_model (
-          OSSO_ABOOK_TREE_VIEW (data->contact_view)));
-  hildon_live_search_widget_hook (HILDON_LIVE_SEARCH (data->live_search),
+          OSSO_ABOOK_TREE_VIEW (main_context->contact_view)));
+  hildon_live_search_widget_hook (
+      HILDON_LIVE_SEARCH (main_context->live_search),
       data->contact_window, GTK_WIDGET (tree_view));
-  gtk_box_pack_start (GTK_BOX (vbox), data->live_search, FALSE, FALSE, 0);
-  gtk_widget_hide (data->live_search);
+  gtk_box_pack_start (GTK_BOX (vbox), main_context->live_search, FALSE, FALSE,
+      0);
+  gtk_widget_hide (main_context->live_search);
 
   gtk_container_add (GTK_CONTAINER (data->contact_window), vbox);
   gtk_widget_show (vbox);
@@ -613,24 +815,28 @@ main (int argc,
     gtk_widget_destroy (data->contact_window);
   g_object_unref (data->handler);
 
-  if (data->selected_master_contact != NULL)
-    g_object_unref (data->selected_master_contact);
+  if (main_context->selected_master_contact != NULL)
+    g_object_unref (main_context->selected_master_contact);
 
-  if (data->selected_contact != NULL)
-    g_object_unref (data->selected_contact);
+  if (main_context->selected_contact != NULL)
+    g_object_unref (main_context->selected_contact);
 
-  if (data->selected_file != NULL)
-    g_object_unref (data->selected_file);
+  if (main_context->selected_file != NULL)
+    g_object_unref (main_context->selected_file);
 
   if (data->notify != NULL)
     g_object_unref (data->notify);
 
+  if (data->send != NULL)
+    g_object_unref (data->send);
+
   if (data->transfers_data != NULL)
     transfers_data_free ((TransfersData *) data->transfers_data);
 
   g_object_unref (unique_app);
 
   g_slice_free (MonorailData, data);
+  g_slice_free (MonorailSendContext, main_context);
 
   return 0;
 }
