@@ -1,7 +1,7 @@
 /*
  * transfer.c
  *
- * Copyright (C) 2009 Collabora Ltd. <http://www.collabora.co.uk/>
+ * Copyright (C) 2009-2010 Collabora Ltd. <http://www.collabora.co.uk/>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -19,6 +19,14 @@
  */
 
 #include "transfer.h"
+
+#include <hildon-mime.h>
+#include <dbus/dbus.h>
+#include <dbus/dbus-glib.h>
+#include <dbus/dbus-gtype-specialized.h>
+#include <dbus/dbus-glib-lowlevel.h>
+#include <hildon/hildon.h>
+
 #include "debug.h"
 
 enum
@@ -30,9 +38,19 @@ enum
   COL_SIGNAL_DONE_ID,
   COL_SIGNAL_ERROR_ID,
   COL_FT_HANDLER,
+  COL_GFILE,
+  COL_STATUS,
   NUM_COLS
 };
 
+enum
+{
+  STATUS_PROGRESS = 0,
+  STATUS_DONE,
+  STATUS_CANCELED,
+  STATUS_ERROR
+};
+
 typedef struct
 {
   EmpathyFTHandler *ft_handler;
@@ -112,49 +130,39 @@ show_or_hide_view (TransfersData *data)
     }
 }
 
-static gboolean
-tap_and_hold_query_cb (GtkWidget *view,
-    GdkEvent *event,
-    TransfersData *data)
-{
-  /* omg */
-  data->x = ((GdkEventButton *)event)->x;
-  data->y = ((GdkEventButton *)event)->y;
-  return FALSE;
-}
-
 static void
-cancel_activate_cb (GtkMenuItem *item,
+cancel_clicked_cb (GtkWidget *b G_GNUC_UNUSED,
     TransfersData *data)
 {
-  GtkTreePath *path;
+  GtkTreeSelection *selection;
+  gboolean selected;
   GtkTreeIter iter;
   EmpathyFTHandler *handler;
   gulong done, error, progress;
   gchar *text, *basename;
+  guint status;
 
-  if (!gtk_tree_view_get_dest_row_at_pos (GTK_TREE_VIEW (data->view),
-          data->x, data->y, &path, NULL))
+  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (data->view));
+  selected = gtk_tree_selection_get_selected (selection, NULL, &iter);
+  if (!selected)
     {
-      DEBUG ("Failed to get path");
+      hildon_banner_show_information (data->view, NULL, "No transfer selected");
       return;
     }
-
-  if (!gtk_tree_model_get_iter (GTK_TREE_MODEL (data->store), &iter, path))
-    {
-      DEBUG ("Failed to get iter");
-      return;
-    }
-
   gtk_tree_model_get (GTK_TREE_MODEL (data->store), &iter,
       COL_FT_HANDLER, &handler,
       COL_SIGNAL_PROGRESS_ID, &progress,
       COL_SIGNAL_DONE_ID, &done,
       COL_SIGNAL_ERROR_ID, &error,
+      COL_STATUS, &status,
       -1);
 
-  if (handler == NULL)
-    return;
+  if (status != STATUS_PROGRESS)
+    {
+      hildon_banner_show_information (data->view, NULL,
+          "Selected transfer is not in progress");
+      return;
+    }
 
   g_signal_handler_disconnect (handler, progress);
   g_signal_handler_disconnect (handler, done);
@@ -173,6 +181,7 @@ cancel_activate_cb (GtkMenuItem *item,
       COL_SIGNAL_DONE_ID, 0,
       COL_SIGNAL_ERROR_ID, 0,
       COL_FT_HANDLER, NULL,
+      COL_STATUS, STATUS_CANCELED,
       -1);
 
   g_free (text);
@@ -185,25 +194,90 @@ cancel_activate_cb (GtkMenuItem *item,
   monorail_transfers_dec (data->monorail_data, TRUE);
 }
 
+static void
+open_clicked_cb (GtkWidget *button G_GNUC_UNUSED,
+    TransfersData *data)
+{
+  GtkTreeSelection *selection;
+  gboolean selected;
+  GtkTreeIter iter;
+  GFile *file;
+  gchar *uri;
+  GSList uri_list;
+  DBusGConnection *connection = dbus_g_bus_get (DBUS_BUS_SESSION, NULL);
+  guint status;
+
+  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (data->view));
+  selected = gtk_tree_selection_get_selected (selection, NULL, &iter);
+  if (!selected)
+    {
+      hildon_banner_show_information (data->view, NULL, "No transfer selected");
+      return;
+    }
+
+  gtk_tree_model_get (GTK_TREE_MODEL (data->store), &iter,
+      COL_GFILE, &file,
+      COL_STATUS, &status,
+      -1);
+
+  if (status != STATUS_DONE)
+    {
+      hildon_banner_show_information (data->view, NULL,
+          "Selected transfer is not finished");
+      return;
+    }
+
+  gtk_widget_hide (data->dialog);
+  data->dialog_visible = FALSE;
+  monorail_transfers_dec (data->monorail_data, FALSE);
+
+  uri = g_file_get_uri (file);
+  uri_list.data = uri;
+  uri_list.next = NULL;
+  hildon_mime_open_file_list (
+    dbus_g_connection_get_connection (connection),
+    &uri_list);
+  g_free (uri);
+}
+
+static void
+clear_clicked_cb (GtkWidget *button G_GNUC_UNUSED,
+    TransfersData *data)
+{
+  GtkTreeIter iter;
+  gboolean valid;
+
+  valid = gtk_tree_model_get_iter_first (GTK_TREE_MODEL (data->store), &iter);
+  while (valid)
+    {
+      guint status;
+      gtk_tree_model_get (GTK_TREE_MODEL (data->store), &iter,
+          COL_STATUS, &status,
+          -1);
+
+      if (status != STATUS_PROGRESS)
+        {
+          valid = gtk_list_store_remove (data->store, &iter);
+          data->num_transfers--;
+          monorail_transfers_dec (data->monorail_data, FALSE);
+        }
+      else
+        {
+          valid = gtk_tree_model_iter_next (GTK_TREE_MODEL (data->store),
+              &iter);
+        }
+    }
+  show_or_hide_view (data);
+}
+
 static GtkWidget *
 create_tree_view (TransfersData *data)
 {
   GtkWidget *view;
+  GtkTreeSelection *selection;
   GtkCellRenderer *renderer;
-  GtkWidget *menu, *child;
 
-  view = hildon_gtk_tree_view_new (HILDON_UI_MODE_NORMAL);
-
-  menu = gtk_menu_new ();
-  child = gtk_menu_item_new_with_label ("Cancel");
-  g_signal_connect (child, "activate", G_CALLBACK (cancel_activate_cb), data);
-  gtk_menu_shell_append (GTK_MENU_SHELL (menu), child);
-  gtk_widget_show (child);
-
-  gtk_widget_tap_and_hold_setup (GTK_WIDGET (view), menu, NULL,
-      GTK_TAP_AND_HOLD_NONE);
-  g_signal_connect (view, "tap-and-hold-query",
-      G_CALLBACK (tap_and_hold_query_cb), data);
+  view = hildon_gtk_tree_view_new (HILDON_UI_MODE_EDIT);
 
   /* Progress */
   renderer = gtk_cell_renderer_progress_new ();
@@ -236,6 +310,9 @@ create_tree_view (TransfersData *data)
 
   gtk_tree_view_set_model (GTK_TREE_VIEW (view), GTK_TREE_MODEL (data->store));
 
+  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (view));
+  gtk_tree_selection_set_mode (selection, GTK_SELECTION_BROWSE);
+
   gtk_widget_show (view);
 
   return view;
@@ -244,9 +321,10 @@ create_tree_view (TransfersData *data)
 static gboolean
 dialog_delete_event_cb (GtkWidget *widget,
     GdkEvent *event,
-    gpointer user_data)
+    TransfersData *data)
 {
   gtk_widget_hide (widget);
+  monorail_transfers_dec (data->monorail_data, FALSE);
   return TRUE;
 }
 
@@ -255,6 +333,10 @@ create_transfer_window (MonorailData *monorail_data)
 {
   TransfersData *data;
   GtkWidget *vbox;
+  GtkWidget *action_area;
+  GtkWidget *open_button;
+  GtkWidget *cancel_button;
+  GtkWidget *clear_button;
 
   data = g_slice_new0 (TransfersData);
   data->monorail_data = monorail_data;
@@ -265,18 +347,51 @@ create_transfer_window (MonorailData *monorail_data)
 
   data->store = gtk_list_store_new (NUM_COLS, G_TYPE_INT, G_TYPE_STRING,
       G_TYPE_STRING, G_TYPE_ULONG, G_TYPE_ULONG, G_TYPE_ULONG,
-      EMPATHY_TYPE_FT_HANDLER);
+      EMPATHY_TYPE_FT_HANDLER, G_TYPE_FILE, G_TYPE_UINT);
 
   data->dialog = gtk_dialog_new ();
   gtk_window_set_title (GTK_WINDOW (data->dialog), "File Transfers");
   gtk_container_set_border_width (GTK_CONTAINER (data->dialog), 5);
   gtk_widget_show (data->dialog);
+  data->dialog_visible = TRUE;
+
+  /* avoid to quit Monorail while the dialog is shown */
+  monorail_transfers_inc (monorail_data, FALSE);
 
   gtk_window_set_transient_for (GTK_WINDOW (data->dialog),
       GTK_WINDOW (monorail_data->contact_window));
 
   g_signal_connect (data->dialog, "delete-event",
-      G_CALLBACK (dialog_delete_event_cb), NULL);
+      G_CALLBACK (dialog_delete_event_cb), data);
+
+  action_area = gtk_dialog_get_action_area (GTK_DIALOG (data->dialog));
+
+  open_button = hildon_button_new_with_text (HILDON_SIZE_FINGER_HEIGHT,
+      HILDON_BUTTON_ARRANGEMENT_VERTICAL,
+      "Open",
+      NULL);
+  g_signal_connect (open_button, "clicked",
+      G_CALLBACK (open_clicked_cb), data);
+  gtk_box_pack_start (GTK_BOX (action_area), open_button, TRUE, TRUE, 0);
+  gtk_widget_show (open_button);
+
+  cancel_button = hildon_button_new_with_text (HILDON_SIZE_FINGER_HEIGHT,
+      HILDON_BUTTON_ARRANGEMENT_VERTICAL,
+      "Cancel",
+      NULL);
+  g_signal_connect (cancel_button, "clicked",
+      G_CALLBACK (cancel_clicked_cb), data);
+  gtk_box_pack_start (GTK_BOX (action_area), cancel_button, TRUE, TRUE, 0);
+  gtk_widget_show (cancel_button);
+
+  clear_button = hildon_button_new_with_text (HILDON_SIZE_FINGER_HEIGHT,
+      HILDON_BUTTON_ARRANGEMENT_VERTICAL,
+      "Clear",
+      NULL);
+  g_signal_connect (clear_button, "clicked",
+      G_CALLBACK (clear_clicked_cb), data);
+  gtk_box_pack_end (GTK_BOX (action_area), clear_button, TRUE, TRUE, 0);
+  gtk_widget_show (clear_button);
 
   vbox = gtk_dialog_get_content_area (GTK_DIALOG (data->dialog));
 
@@ -378,6 +493,7 @@ transfer_progress (EmpathyFTHandler *handler,
       COL_PROGRESS_PERCENT, percentage,
       COL_TRANSFER_STRING, text,
       COL_REMAINING_STRING, remaining,
+      COL_STATUS, STATUS_PROGRESS,
       -1);
 
   g_free (remaining);
@@ -423,6 +539,7 @@ transfer_done (EmpathyFTHandler *handler,
       COL_SIGNAL_DONE_ID, 0,
       COL_SIGNAL_ERROR_ID, 0,
       COL_FT_HANDLER, NULL,
+      COL_STATUS, STATUS_DONE,
       -1);
 
   g_free (text);
@@ -473,6 +590,7 @@ transfer_error (EmpathyFTHandler *handler,
       COL_SIGNAL_DONE_ID, 0,
       COL_SIGNAL_ERROR_ID, 0,
       COL_FT_HANDLER, NULL,
+      COL_STATUS, STATUS_ERROR,
       -1);
 
   g_free (text);
@@ -489,6 +607,7 @@ transfer_new_handler (MonorailHandler *handler,
   TransfersData *data;
   GtkTreeIter iter;
   gchar *text, *basename;
+  GFile *file;
 
   monorail_transfers_inc (monorail_data, TRUE);
 
@@ -506,7 +625,8 @@ transfer_new_handler (MonorailHandler *handler,
   error = g_signal_connect (ft_handler, "transfer-error",
       G_CALLBACK (transfer_error), data);
 
-  basename = g_file_get_basename (empathy_ft_handler_get_gfile (ft_handler));
+  file = g_file_dup (empathy_ft_handler_get_gfile (ft_handler));
+  basename = g_file_get_basename (file);
   text = g_strdup_printf ("%s %s\n%s",
       empathy_ft_handler_is_incoming (ft_handler) ? "Receiving" : "Sending",
       basename,
@@ -514,6 +634,7 @@ transfer_new_handler (MonorailHandler *handler,
   g_free (basename);
 
   data->num_transfers++;
+  monorail_transfers_inc (monorail_data, FALSE);
   gtk_list_store_append (data->store, &iter);
   gtk_list_store_set (data->store, &iter,
       COL_PROGRESS_PERCENT, 0,
@@ -523,6 +644,8 @@ transfer_new_handler (MonorailHandler *handler,
       COL_SIGNAL_DONE_ID, done,
       COL_SIGNAL_ERROR_ID, error,
       COL_FT_HANDLER, ft_handler,
+      COL_GFILE, file,
+      COL_STATUS, STATUS_PROGRESS,
       -1);
 
   show_or_hide_view (data);
@@ -535,9 +658,17 @@ transfer_new_handler (MonorailHandler *handler,
 void
 transfer_show_window (MonorailData *monorail_data)
 {
+  TransfersData *data = (TransfersData *) monorail_data->transfers_data;
+
   if (monorail_data->transfers_data == NULL)
     create_transfer_window (monorail_data);
+  else 
+    {
+      if (!data->dialog_visible)
+        monorail_transfers_inc (monorail_data, FALSE);
+    }
 
   gtk_window_present (
       GTK_WINDOW (((TransfersData *) monorail_data->transfers_data)->dialog));
+  data->dialog_visible = TRUE;
 }
